(function() {
  var CoffeeQWorker, EventEmitter, redis;
  var __hasProp = Object.prototype.hasOwnProperty, __extends = function(child, parent) {
    for (var key in parent) { if (__hasProp.call(parent, key)) child[key] = parent[key]; }
    function ctor() { this.constructor = child; }
    ctor.prototype = parent.prototype;
    child.prototype = new ctor;
    child.__super__ = parent.prototype;
    return child;
  }, __bind = function(fn, me){ return function(){ return fn.apply(me, arguments); }; }, __slice = Array.prototype.slice;
  redis = require('redis');
  EventEmitter = require('events').EventEmitter;
  /*
  CoffeeQWorker
  ============
  Uses two redis client instances so that it can handle pubsub and normal events simultaneously (redis connections have a pubsub mode that prevents one from doing both pubsub and normal events simultaneously)
  @queueClient, @pubsubClient
  
  
  Events:
  =======
  Emits 'message' each time a pubsub message from Redis is received.
    err    - The caught exception.
    channel - The pubsub channel that the message is received on
    message  - The message data
  
  Emits 'job' before attempting to run any job.
    worker - This Worker instance.
    queue  - The String queue that is being checked.
    job    - The parsed Job object that was being run.
  
  Emits 'success' after a successful job completion.
    worker - This Worker instance.
    queue  - The String queue that is being checked.
    job    - The parsed Job object that was being run.
    result - The result produced by the job.
  
  Emits 'error' if there is an error fetching or running the job.
    err    - The caught exception.
    worker - This Worker instance.
    queue  - The String queue that is being checked.
    job    - The parsed Job object that was being run.
  */
  CoffeeQWorker = (function() {
    function CoffeeQWorker(queue, callbacks, options) {
      if (!options) {
        options = {};
      }
      this.port = options.port || 6379;
      this.host = options.host || 'localhost';
      this.queue = queue;
      this.queue_key = this.key('queue', queue);
      this.callbacks = callbacks || {};
      this.queueClient = redis.createClient(this.port, this.host);
      this.pubsubClient = redis.createClient(this.port, this.host);
      this.registerMessageHandlers(this.queue_key);
    }
    __extends(CoffeeQWorker, EventEmitter);
    CoffeeQWorker.prototype.start = function() {
      console.log("start worker");
      this.clearQueue(this.queue_key);
      return this.subscribeToQueueChannel(this.queue_key);
    };
    CoffeeQWorker.prototype.stop = function() {
      console.log("end worker");
      return this.disconnectFromQueueChannel(this.queue_key);
    };
    CoffeeQWorker.prototype.subscribeToQueueChannel = function(channel) {
      console.log("subscribe to " + channel);
      return this.pubsubClient.subscribe(channel);
    };
    CoffeeQWorker.prototype.disconnectFromQueueChannel = function(channel) {
      console.log("unsubscribe from " + channel);
      return this.pubsubClient.end();
    };
    CoffeeQWorker.prototype.registerMessageHandlers = function(channel) {
      this.pubsubClient.on("message", __bind(function(channel, message) {
        console.log("client 1 channel:" + channel + " message:" + message);
        this.emit('message', this, channel, message);
        if (message === "queued") {
          return this.popAndRun(channel);
        }
      }, this));
      return this.pubsubClient.on("subscribe", function(channel, count) {
        return console.log("subscribed to " + channel);
      });
    };
    CoffeeQWorker.prototype.clearQueue = function(queue) {
      return this.queueClient.llen(queue, __bind(function(err, resp) {
        console.log("CLEAR QUEUE resp: " + resp);
        if (resp > 0) {
          this.popAndRun(queue);
          return this.clearQueue(queue);
        }
      }, this));
    };
    /*
      Pulls the job off the queue and executes it  
      returns nothing
      */
    CoffeeQWorker.prototype.popAndRun = function(queue) {
      return this.queueClient.lpop(queue, __bind(function(err, resp) {
        if (!err && resp) {
          console.log("resp " + resp);
          return this.perform(JSON.parse(resp.toString()));
        } else {
          return console.log("Error popping " + err);
        }
      }, this));
    };
    /*
      Handles the actual running of the job.    
      job - The parsed Job object that is being run.    
      Returns nothing.
      */
    CoffeeQWorker.prototype.perform = function(job) {
      var cb, old_title;
      old_title = process.title;
      this.emit('job', this, this.queue, job);
      this.procline("" + this.queue + " job since " + ((new Date).toString()));
      if (cb = this.callbacks[job["class"]]) {
        return cb.apply(null, __slice.call(job.args).concat([__bind(function(result) {
          try {
            if (result instanceof Error) {
              return this.fail(result, job);
            } else {
              return this.succeed(result, job);
            }
          } finally {
            console.log("done performing");
          }
        }, this)]));
      } else {
        return this.fail(new Error("Missing Job: " + job["class"]), job);
      }
    };
    /*
      Tracks stats for successfully completed jobs.    
      result - The result produced by the job. 
      job    - The parsed Job object that is being run.    
      Returns nothing.
      */
    CoffeeQWorker.prototype.succeed = function(result, job) {
      return this.emit('success', this, this.queue, job, result);
    };
    /*
      Tracks stats for failed jobs, and tracks them in a Redis list.  
      err - The caught Exception.
      job - The parsed Job object that is being run.
      Returns nothing.
      */
    CoffeeQWorker.prototype.fail = function(err, job) {
      return this.emit('error', err, this, this.queue, job);
    };
    /*
      Sets the process title.    
      msg - The String message for the title.    
      Returns nothing.
      */
    CoffeeQWorker.prototype.procline = function(msg) {
      return process.title = "resque-" + exports.version + ": " + msg;
    };
    Object.defineProperty(CoffeeQWorker.prototype, 'name', {
      get: function() {
        return this._name;
      },
      set: function(name) {
        return this._name = this.ready ? [name || 'node', process.pid, this.queues].join(":") : name;
      }
    });
    /*
      Builds a namespaced Redis key with the given arguments.    
      args - Array of Strings.    
      Returns an assembled String key.
      */
    CoffeeQWorker.prototype.key = function() {
      var args;
      args = 1 <= arguments.length ? __slice.call(arguments, 0) : [];
      args.unshift(this.namespace);
      return args.join(":");
    };
    return CoffeeQWorker;
  })();
  module.exports = CoffeeQWorker;
}).call(this);
